# 参照地表水GB3838-2002
CLASS_LIMITS = {
    1: {
        'DO': 7.5,  # 溶解氧，mg/L
        'CODMn': 2.0,  # 高锰酸盐指数，mg/L
        'NH3_N': 0.15,  # 氨氮，mg/L
        'TN': 0.2,  # 总氮，mg/L
        'TP': 0.02,  # 总磷，mg/L。湖、库I-V限值0.01/0.025/0.05/0.1/0.2
    },
    2: {
        'DO': 6.0,
        'CODMn': 4.0,
        'NH3_N': 0.5,
        'TN': 0.5,
        'TP': 0.1,
    },
    3: {
        'DO': 5.0,
        'CODMn': 6.0,
        'NH3_N': 1.0,
        'TN': 1.0,
        'TP': 0.2,
    },
    4: {
        'DO': 3.0,
        'CODMn': 10.0,
        'NH3_N': 1.5,
        'TN': 1.5,
        'TP': 0.3,
    },
    5: {
        'DO': 2.0,
        'CODMn': 15.0,
        'NH3_N': 2.0,
        'TN': 2.0,
        'TP': 0.4,
    },
}

def classify_water_quality(v: list or float, species:str or list = ['DO','CODMn', 'NH3_N', 'TN', 'TP']):
    def __class2str(c):
        if c <= 3:
            return "优良（I~III），水源地、自然保护区、水产养殖区等"
        elif c==4:
            return "四类，工业用水区及人体非直接接触娱乐用水区"
        elif c==5:
            return "五类，农业用水及一般景观用水"
        else:
            return "劣五类，污染水域"
    if isinstance(v, float):
        if species in CLASS_LIMITS[5].keys():
            if species != "DO":
                for i in range(5):
                    if v <= CLASS_LIMITS[i+1][species]:
                        return i+1, __class2str(i+1)
                return 6, __class2str(6)
            else:
                for i in range(5):
                    if v >= CLASS_LIMITS[i+1][species]:
                        return i+1, __class2str(i+1)
                return 6, __class2str(6)
        else:
            raise Exception(f'species should be set properly, accept: DO, CODMn, NH3_N, TN, TP, given {species}')
    elif isinstance(v, list) and len(v) == 5:
        classes = []
        for x, s in zip(v,species):
            classes.append(classify_water_quality(x, s)[0])
        return min(classes), __class2str(min(classes))
    else:
        raise Exception(f'cannot resolve input value: {v}')



import datetime
import pandas as pd
import numpy as np
from .ml import autoreg_fill
def tryfloat(x, nanflag:float = None, nanvalue=np.nan):
    if nanflag and np.abs(float(x) - nanflag) < 1E-8:
        return nanvalue
    try:
        return float(x)
    except:
        return nanvalue


def moving_avg(x:np.ndarray, w:int) -> np.ndarray:
    if w % 2 != 1:
        raise ValueError(f'moving_avg w % 2 == 1, given {w}')
    data = np.convolve(x, np.ones(w), 'same') / w
    hw = int((w-1)/2)
    data[:hw] = data[hw]
    data[-hw:] = data[-(hw+1)]
    return data


def outlier_filter_single_point(data, thres=3.0):
    xm = np.nanmedian(data)
    xstd = np.nanstd(data)
    N = len(data)
    for i in range(N):
        if not np.isnan(data[i]):
            if i == 0 or np.isnan(data[i - 1]):
                xb = data[i] - xm
            else:
                xb = data[i] - data[i - 1]
            if i == (N - 1) or np.isnan(data[i + 1]):
                xf = data[i] - xm
            else:
                xf = data[i] - data[i + 1]
            if xb * xf > 0 and xb > thres * xstd and xf > thres * xstd:
                data[i] = np.nan
    return data


def outlier_filer_sigma(data, thres):
    d = data - moving_avg(data, 9)
    I = np.abs(d - d.mean()) > thres * d.std()
    data[I] = np.nan
    return data

class ExcelDataProcessor(object):
    def __init__(self, fn, sheetname):
        self.df = pd.read_excel(fn, sheetname)

    def try_convert_float(self, key, nanflag=None):
        if key in self.df.keys():
            self.df[key] = self.df[key].map(lambda x: tryfloat(x, nanflag))
        else:
            raise ValueError(f'ExcelDataProcessor: unexpected key {key}')

    def try_convert_datetime(self, key, format):
        if key in self.df.keys():
            self.df[key] = self.df[key].map(lambda x: datetime.datetime.strptime(x, format))
        else:
            raise ValueError(f'ExcelDataProcessor: unexpected key {key}')

    def try_convert_str(self, key):
        if key in self.df.keys():
            self.df[key] = self.df[key].astype(str)
        else:
            raise ValueError(f'ExcelDataProcessor: unexpected key {key}')

    def left_merge_by_key(self, df, key):
        self.df = pd.merge(left=self.df, right=df, how='left', on=key)

    def outlier_filter(self, key, thres=3.0):
        if key in self.df.keys():
            # plt.plot(self.df[key].to_numpy())
            # plt.plot(outlier_filer_sigma(self.df[key].interpolate().to_numpy(), thres))
            # plt.title(key)
            # plt.show()
            self.df[key] = pd.Series(outlier_filer_sigma(self.df[key].interpolate().to_numpy(), thres))
        else:
            raise ValueError(f'ExcelDataProcessor: unexpected key {key}')

    def autoreg_fill(self, key, rank):
        if key in self.df.keys():
            self.df[key] = pd.Series(autoreg_fill(self.df[key].to_numpy(), rank))
        else:
            raise ValueError(f'ExcelDataProcessor: unexpected key {key}')







def load_config(path='config', encoding='utf-8'):
    def __tryformat(s):
        try:
            return int(s)
        except:
            try:
                return float(s)
            except:
                return s

    config = {}
    with open(path, 'r', encoding=encoding) as f:
        for l in f.readlines():
            if l.startswith('==='):
                continue
            else:
                newl = l
                if '&&&' in l:
                    newl = l.split('&&&')[0].strip()
                if '=' in newl:
                    keyvalue = newl.split('=')
                    keyvalue = [x.strip() for x in keyvalue]
                    if len(keyvalue) != 2:
                        print(f'WARNING: config file {path} has line {l} with invalid format')
                    config[keyvalue[0]] = __tryformat(keyvalue[1])
                else:
                    print(f'WARNING: config file {path} has line {l} with invalid format')
    return config


import pymysql
import pymongo
import urllib


class MySQLConn(object):
    def __init__(self, host: str, user: str, password: str, db: str, port: int):
        self.conn = pymysql.connect(host=host, user=user, password=password, db=db, port=port)
        self.cursor = self.conn.cursor()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.conn.close()

    def excute_one(self, comm):
        self.cursor.execute(comm)
        res = self.cursor.fetchone()
        if res is None:
            return None
        elif len(res) == 1:
            return res[0]
        else:
            return res

    def excute_all(self, comm):
        self.cursor.execute(comm)
        return self.cursor.fetchall()

    def insert_confirm(self, comm):
        x = self.cursor.execute(comm)
        if x > 0:
            self.conn.commit()
        else:
            self.conn.rollback()
        return x

    def insert(self, comm):
        return self.cursor.execute(comm)

    def confirm(self):
        self.conn.commit()

    def rollback(self):
        self.conn.rollback()


class MongoConn(object):
    def __init__(self, user:str, password:str, host:str, port:int, db:str, authMechanism:str='DEFAULT', authSource:str=None):
        if authSource is None:
            self.conn = pymongo.MongoClient(f'mongodb://{urllib.parse.quote_plus(user)}:{urllib.parse.quote_plus(password)}@{host}:{port}'
                                            f'/?authMechanism={authMechanism}')
        else:
            self.conn = pymongo.MongoClient(f'mongodb://{urllib.parse.quote_plus(user)}:{urllib.parse.quote_plus(password)}@{host}:{port}'
                                            f'/?authMechanism={authMechanism}&authSource={authSource}')
        self.cursor = self.conn[db]

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.conn.close()

    def fetch_one(self, tablename:str, filter: dict):
        return self.cursor[tablename].find_one(filter)

    def update_one(self, tablename:str, filter:dict, data:dict):
        x = self.fetch_one(tablename, filter)
        if x:
            return self.cursor[tablename].update_one(filter, {'$set': data})
        else:
            return None

    def insert_one(self, tablename:str, data:dict):
        self.cursor[tablename].insert_one(data)





import os
import numpy as np
from haversine import haversine
import matplotlib
import matplotlib.pyplot as plt
import pickle

matplotlib.use('Agg')

class SimpleDisWeightModel(object):
    def __init__(self, x, y, z, e=0.2):
        self.x = x
        self.y = y
        self.z = z
        self.e = e

    def excute(self, gridx, gridy):
        X, Y = np.meshgrid(gridx, gridy)
        total_weight = np.zeros_like(X)
        total_value = np.zeros_like(X)
        for i in range(len(self)):
            w = 1 / np.power(np.power(X - self.x[i], 2) + np.power(Y - self.y[i], 2), self.e)
            total_weight += w
            total_value += w * self.z[i]
        return np.divide(total_value, total_weight)

    def __len__(self):
        return len(self.x)

def rgba_to_hex(rgba):
    r, g, b, a = rgba
    return "#{:02x}{:02x}{:02x}{:02x}".format(int(r * 255), int(g * 255), int(b * 255), int(a * 255))

def heatmap(X:np.ndarray, Y:np.ndarray, Q:np.ndarray, combine=None, combine_method='sum', imgpath=None, resolution=512,
            offset=0.1, levels=None, cache=None, cmap='rainbow')->str:
    def __get_nearest_points(X: np.ndarray, Y:np.ndarray):
        n = len(X)
        mini = 0
        minj = 1
        minV = 1e10
        for i in range(n - 1):
            for j in range(i + 1, n):
                dist = haversine((Y[i], X[i]), (Y[j], X[j])) * 1000
                if dist < minV:
                    mini, minj = i, j
                    minV = dist
        return mini, minj, minV

    def __combine_points(i, j, X, Y, Q, method):
        X[i] = (X[i] + X[j])/2
        Y[i] = (Y[i] + Y[j])/2
        if method == 'sum':
            Q[i] = Q[i] + Q[j]
        elif method == 'avg':
            Q[i] = (Q[i] + Q[j])/2
        else:
            raise ValueError(f'only accept method sum and avg, given {method}')
        return np.delete(X, j), np.delete(Y, j), np.delete(Q, j)

    def __save_cache(path, X, Y, Q):
        with open(path, 'wb') as f:
            pickle.dump({'X': X, 'Y': Y, 'Q': Q}, f)

    def __load_cache(path):
        with open(path, 'rb') as f:
            r = pickle.load(f)
        return r['X'], r['Y'], r['Q']

    if combine is not None:
        mini, minj, minV = __get_nearest_points(X, Y)
        while(minV <= combine):
            X, Y, Q = __combine_points(mini, minj, X, Y, Q, combine_method)
            print(minV, len(X))
            mini, minj, minV = __get_nearest_points(X, Y)
    # print('s')
    if cache is not None:
        if os.path.exists(cache):
            X, Y, Q = __load_cache(cache)
        else:
            __save_cache(cache, X, Y, Q)

    xstart, xend = min(X)-offset, max(X)+offset
    ystart, yend = min(Y)-offset, max(Y)+offset
    gridx = np.arange(xstart, xend, (xend - xstart) / resolution)
    gridy = np.arange(ystart, yend, (yend - ystart) / resolution)
    m = SimpleDisWeightModel(X, Y, Q, e=1.0)
    z = m.excute(gridx, gridy)
    plt.figure(figsize=(10, 10 * (yend - ystart) / (yend - ystart)))
    # info = plt.contourf(gridx, gridy, z, cmap='rainbow', extend='both')
    print(f'{z.max():.4f}, {z.min():.4f}')
    # info = plt.contourf(gridx, gridy, z, levels=np.linspace(info.levels[0], info.levels[-1], 8), cmap='rainbow',
    #                     extend='both')
    if levels is None:
        info = plt.contourf(gridx, gridy, z, levels=np.percentile(z.flatten(), np.linspace(100/8, 700/8, num=8)), cmap=cmap,
                            extend='both')
    else:
        info = plt.contourf(gridx, gridy, z, levels=levels, cmap=cmap, extend='both')
    # plt.colorbar(info)
    # plt.show()
    plt.axis('off')
    if imgpath is None:
        imgpath = 'heatmap.png'
    plt.savefig(imgpath, bbox_inches='tight', pad_inches=0, transparent=True)  # https://www.zhihu.com/question/506015285/answer/2270700851
    plt.close()
    hexcode = [rgba_to_hex(x) for x in info.get_cmap()(np.linspace(0, 1, len(info.levels)))]
    return imgpath, [xstart, ystart], [xend, yend], info.levels.tolist(), hexcode


from setuptools import setup, find_packages

with open('README.md', 'r') as f:
    long_des = f.read()

setup(
    name='jsepp',
    version='0.1.8',
    packages=find_packages(),
    install_requires=[
        'pymysql',
        'pymongo',
        'numpy',
        'pandas',
        'openpyxl',
        'scipy',
        'statsmodels',
        'haversine'
    ],
    author='Hansen Zhao',
    author_email='zhaohs12@163.com',
    description='a package for general environmental data processing',
    long_description=long_des
)


from statsmodels.tsa.ar_model import AutoReg
import numpy as np

def autoreg_fill(x:np.ndarray, rank:int) -> np.ndarray:
    for i in range(len(x)):
        if np.isnan(x[i]):
            if i > 10:
                model = AutoReg(x[:i], rank)
                r = model.fit()
                x[i] = r.predict(i, i)[0]
            elif i == 0:
                for j in range(len(x)):
                    if not np.isnan(x[j]):
                        x[0] = x[j]
                        break
            else:
                x[i] = x[i - 1]
    return x